// Implementazione dei fattori primi 

#include "vlong.hpp"
#include "prime.hpp"

int is_probable_prime( const vlong &p )
{
  // Effettuiamo un test basato sul teorema del matematico P. Fermat (vedi articolo):
  // a**(p-1) = 1 mod p 
  // per il numero primo p
  // per numeri da 1000bit questa operazione puo' richiedere molto tempo
  // sara' quindi necessario migliorare la funzione e/o rivederla completamente,
  // in una successiva implementazione.
  const rep = 4;
  const unsigned any[rep] = { 2,3,5,7 /*,11,13,17,19,23,29,31,37..*/ };
  for ( unsigned i=0; i<rep; i+=1 )
    if ( modexp( any[i], p-1, p ) != 1 )
      return 0;
  return 1;
}

prime_factory::prime_factory( unsigned MP )
{
  np = 0;

  // Iniziamo!, Inizializziamo...
  char * b = new char[MP+1]; // Carattere extra, per fermare la ricerca
  for (unsigned i=0;i<=MP;i+=1) b[i] = 1;
  unsigned p = 2;
  while (1)
  {
    // saltiamo i compositi
    while ( b[p] == 0 ) p += 1;
    if ( p == MP ) break;
    np += 1;
    // rimuoviamo i multipli
    unsigned c = p*2;
    while ( c < MP )
    {
      b[c] = 0;
      c += p;
    }
    p += 1;
  }
  pl = new unsigned[np];
  np = 0; for (p=2;p<MP;p+=1) if ( b[p] ) { pl[np] = p; np += 1; }
  delete [] b;
}

prime_factory::~prime_factory()
{
  delete [] pl;
}

vlong prime_factory::find_prime( vlong & start )
{
  unsigned SS = 1000; // in generale, dovrebbe bastare (fortuna permettendo!)
  char * b = new char[SS]; // bitset dei primi candidati
  unsigned tested = 0;
  while (1)
  {
    unsigned i;
    for (i=0;i<SS;i+=1)
      b[i] = 1;
    for (i=0;i<np;i+=1)
    {
      unsigned p = pl[i];
      unsigned r = to_unsigned(start % p); // questo e' un punto cruciale.
	  // non e' veloce come dovrebbe. In effetti, andra' reimplementato con routines dedicate.
      if (r) r = p - r;
      // eliminiamo i multipli di p
      while ( r < SS )
      {
        b[r] = 0;
        r += p;
      }
    }
  // adesso, testiamo i nostri candidati.
	// Una buona dose di fortuna, ed avremo i nostri primi!
    for (i=0;i<SS;i+=1)
    {
      if ( b[i] )
      {
        tested += 1;
        if ( is_probable_prime(start) )
        {
          delete [] b;
          return start;
        }
      }
      start += 1;
    }
  }
}

int prime_factory::make_prime( vlong & r, vlong &k, const vlong & min_p )
{
  k = 1;
  for (unsigned i=0;i<np;i+=1)
  {
     unsigned p = pl[i];
     while ( r % p == 0 )
     {
       if ( r == p )
         return 1; // teoricamente, puo' succedere solo se min_p e' piccolo.
       r = r / p;
       k = k * p;
       if ( r < min_p )
         return 0;
     }
  }
  return is_probable_prime( r );
}
