class flex_unit // spazio per il salvataggio e controlli sull'indice
{
  unsigned * a; // array unita'
  unsigned z; // unita' allocate
public:
  unsigned n; // unita' usate (solo lettura)
  flex_unit();
  ~flex_unit();
  void clear(); // settiamo n a zero
  unsigned get( unsigned i ) const;     // prendiamo i unsigned
  void set( unsigned i, unsigned x );   // settiamo i unsigned
  void reserve( unsigned x );           // salviamo l'elemento

  // Routime Critica - Da controllare ed ottimizzare
  void fast_mul( flex_unit &x, flex_unit &y, unsigned n );
};

class vlong_value : public flex_unit
{
  public:
  unsigned share; // contatore condiviso, usato da vlonng per rallentamenti
  int is_zero() const;
  unsigned bit( unsigned i ) const;
  void setbit( unsigned i );
  void clearbit( unsigned i );
  unsigned bits() const;
  int cf( vlong_value& x ) const;
  int product( vlong_value &x ) const;
  void shl();
  int  shr(); // riporto 
  void shr( unsigned n );
  void add( vlong_value& x );
  void xor( vlong_value& x );
  void and( vlong_value& x );
  void subtract( vlong_value& x );
  void init( unsigned x );
  void copy( vlong_value& x );
  unsigned to_unsigned(); // Conversione Non ottimizzata! Prima cosa da fixare!
  vlong_value();
  void mul( vlong_value& x, vlong_value& y );
  void divide( vlong_value& x, vlong_value& y, vlong_value& rem );
};

class vlong // very long integer - eheheh, adesso possiamo usarlo come long!...r00x!
{
public:
  // Reimplementazione degli operatori aritmetici...
  friend vlong operator +( const vlong& x, const vlong& y );
  friend vlong operator -( const vlong& x, const vlong& y );
  friend vlong operator *( const vlong& x, const vlong& y );
  friend vlong operator /( const vlong& x, const vlong& y );
  friend vlong operator %( const vlong& x, const vlong& y );
  friend vlong operator ^( const vlong& x, const vlong& y );
  friend vlong pow2( unsigned n );
  friend vlong operator &( const vlong& x, const vlong& y );

  friend vlong operator <<( const vlong& x, unsigned n );

  vlong& operator +=( const vlong& x );
  vlong& operator -=( const vlong& x );
  vlong& operator >>=( unsigned n );

  // Reimplementazione degli operatori di Confronto 
  friend int operator !=( const vlong& x, const vlong& y );
  friend int operator ==( const vlong& x, const vlong& y );
  friend int operator >=( const vlong& x, const vlong& y );
  friend int operator <=( const vlong& x, const vlong& y );
  friend int operator > ( const vlong& x, const vlong& y );
  friend int operator < ( const vlong& x, const vlong& y );

  // Valore Assoluto
  friend vlong abs( const vlong & x );

  // Operazioni di costruzione e conversione
  vlong ( unsigned x=0 );
  vlong ( const vlong& x );
  ~vlong();
  friend unsigned to_unsigned( const vlong &x );
  vlong& operator =(const vlong& x);

  // Lavori sui Bit - molto lavoro di ottimizzazione, ankora da fare
  unsigned bits() const;
  unsigned bit(unsigned i) const;
  void setbit(unsigned i);
  void clearbit(unsigned i);
  vlong& operator ^=( const vlong& x );
  vlong& operator &=( const vlong& x );
  vlong& ror( unsigned n ); // rotazione bit
  vlong& rol( unsigned n ); // rotazione bit
  friend int product( const vlong & x, const vlong & y ); //  = x&y

  void load( unsigned * a, unsigned n ); // caricamento valore, a[0] sara' lsw
  void store( unsigned * a, unsigned n ) const; // salvataggio a basso livello, a[0] - lsw

private:
  class vlong_value * value;
  int negative;
  int cf( const vlong & x ) const;
  void docopy();
  friend class monty;
};

vlong modexp( const vlong & x, const vlong & e, const vlong & m ); // m DEVE essere dispari!!

vlong gcd( const vlong &X, const vlong &Y ); // massimo comune denominatore
vlong modinv( const vlong &a, const vlong &m ); // inverso mod

vlong monty_exp( const vlong & x, const vlong & e, const vlong & m );
vlong monty_exp( const vlong & x, const vlong & e, const vlong & m, const vlong &p, const vlong &q );

class rng
{
public:
  virtual vlong next()=0;
};

class vlstr
{
public:
  virtual void put( const vlong & x )=0;
  virtual vlong get()=0;
};

vlong lucas ( vlong P, vlong Z, vlong k, vlong p ); // P^2 - 4Z != 0
vlong sqrt( vlong g, vlong p ); // radice quadrata mod p
