/************************************************************************
* sls.c			Simple Linux Sniffer v.0.3			*
*									*
*			semplice sniffer di rete in grado di monitorare *
*			pacchetti IP con header TCP, UDP e ICMP.	* 
*			Il codice fa uso del particolare tipo di socket *
*			SOCK_PACKET implementato SOLO su Linux al posto *
*			dei vari NIT, BPF, DLPI dei vari SVr4 e BSD4.4+ *
*			per raggiungere il livello 'datalink'.		*
*			Porta l'interfaccia di rete eth0 in modalita'	*
*			promiscua, permettendole di accogliere anche i	*
*			pacchetti non specificatamente indirizzati ad	*
*			essa. Un semplice signal_catcher rimuove quella *
*			flag IFF_PROMISC in caso di interruzione o kill	*
*			del programma.					*
*									*
* OS:			Linux (SOCK_PACKET)				*
*									*
* 				NO(C)1998 FuSyS TCP/IP Tools Unlimited	*
************************************************************************/

/*
*   SPECIAL TNX TO CODER@REPTILE FOR HIS IDEAS AND WORK
*/ 

#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <linux/if.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>

#define		MTU			1500
#define 	URG                     32
#define		ACK_PSH			24
#define		SYN_ACK			18
#define		FIN_ACK			17
#define 	ACK                     16
#define 	PSH                     8
#define 	RST                     4
#define 	SYN                     2
#define 	FIN                     1
#define		ETH			"eth0"	
#define		ETH_HD			14

int sniff;
int tcp_cn = 1;
int udp_cn = 1;

struct packet_info {
	unsigned char ttl;
	unsigned char protocol;
	unsigned char *saddr, *daddr;
	unsigned long seq, ack_seq;
	unsigned short source, dest;
	unsigned short type, id;
	unsigned short flags;
	unsigned short window;
	char *dataload;
};

struct TCPhdr {
	unsigned short source, dest;
        unsigned long seq, ack_seq;
        unsigned short offset_flag, window, checksum, urgent;
};


int init()
{
	int fd;
	struct ifreq eth;
	if( (fd = socket(AF_INET, SOCK_PACKET, htons(0x3))) < 0) {
		fprintf(stderr, "Can't open RAW Socket.\n");
		exit(1);
	}
	strcpy(eth.ifr_ifrn.ifrn_name, ETH);
	if ( ioctl(fd, SIOCGIFFLAGS, &eth) < 0) {
		fprintf(stderr, "Can't get Ethernet flags.\n");
		exit(1);
	}
	eth.ifr_ifru.ifru_flags |= IFF_PROMISC ;
	if ( ioctl(fd, SIOCSIFFLAGS, &eth) < 0) {
		fprintf(stderr, "Can't put Ethernet in PROMISC mode.\n");
		exit(1);
	}
	if ( fcntl(fd,F_SETOWN, getpid()) < 0) {
		fprintf(stderr, "Can't set SOCK_PACKET Ownership.\n");
		exit(1);
	}
	return fd;
}

int sniff_pk(int fd, struct packet_info *infoz)
{
	int pk_len;
	char sniff_buff[MTU];
	struct iphdr *IP;
	struct TCPhdr *TCP;
	struct udphdr *UDP;
	struct icmphdr *ICMP;
	char data[MTU];

	memset(sniff_buff, '\0', MTU);
        memset(data, '\0', MTU);
	pk_len = read(fd, sniff_buff, MTU);
	if (read > 0) {
		pk_len -= ETH_HD;
		memcpy(data, sniff_buff+ETH_HD, pk_len);
		IP = (struct iphdr *) data;
		infoz->ttl = IP->ttl;
		infoz->protocol = (char)IP->protocol;
		infoz->saddr = (unsigned char *)&(IP->saddr);
		infoz->daddr = (unsigned char *)&(IP->daddr);
		switch (infoz->protocol) {
			case IPPROTO_TCP:
			TCP = (struct TCPhdr *)(data+20);
			infoz->seq = ntohl(TCP->seq);
			infoz->ack_seq = ntohl(TCP->ack_seq);
			infoz->source = ntohs(TCP->source);
			infoz->dest = ntohs(TCP->dest);
			infoz->window = ntohs(TCP->window);
			infoz->flags = ntohs(TCP->offset_flag)&
					(URG|ACK|PSH|FIN|RST|SYN);
			infoz->dataload = (char *)(data +
				(sizeof(struct iphdr)+sizeof(struct TCPhdr)));
			break;
		 	case IPPROTO_UDP:
			UDP = (struct udphdr *)(data+20);
			infoz->source = ntohs(UDP->source);
			infoz->dest = ntohs(UDP->dest);
			infoz->dataload = (char *)(data +
				(sizeof(struct iphdr)+sizeof(struct udphdr)));
			break;
			case IPPROTO_ICMP:
			ICMP = (struct icmphdr *)(data+20);
			infoz->type = ntohs(ICMP->type);
			infoz->id = ntohs(ICMP->un.echo.sequence);
			break;
			default:
			break;
		}
	}
	return pk_len;		
}

void ethclose()
{
	struct ifreq eth;
	strcpy(eth.ifr_ifrn.ifrn_name, ETH);
        if ( ioctl(sniff, SIOCGIFFLAGS, &eth) < 0) {
                fprintf(stderr, "Can't get Ethernet flags.\n");
                exit(1);
        }
        eth.ifr_ifru.ifru_flags ^= IFF_PROMISC ;
        if ( ioctl(sniff, SIOCSIFFLAGS, &eth) < 0) {
                fprintf(stderr, "Can't put Ethernet in PROMISC mode.\n");
                exit(1);	
	}
	exit(0);
}

void dump_tcp(struct packet_info info, int data)
{
	int count = 1;
	char *flags;
	
	printf("\nTCP Packet #%i\n", tcp_cn++);
        printf("%u.%u.%u.%u:%i > %u.%u.%u.%u:%i\tSEQ:%u ACK:%u\n",
         	info.saddr[0],info.saddr[1],info.saddr[2],
                info.saddr[3], info.source, info.daddr[0],
                info.daddr[1], info.daddr[2], info.daddr[3],
                info.dest, info.seq, info.ack_seq);
                printf("(TTL:%i Window:%i)\t\t\t", info.ttl, info.window);

	switch (info.flags) {
		case URG:
                	flags="-----U";
                        break;
                case ACK_PSH:
                        flags="---PA-";
                        break;
                case SYN_ACK:
                        flags="-S--A-";
                        break;
                case FIN_ACK:
                        flags="F---A-";
                        break;
                case ACK:
                        flags="----A-";
                        break;
                case PSH:
                        flags="---P--";
                        break;
                case RST:
                        flags="--R---";
                        break;
                case SYN:
                        flags="-S----";
                        break;
                case FIN:
                        flags="F-----";
			break;
		default:
			break;
	}
	printf("FLAGS: %s\n", flags);
        if (data == 1) {
        	printf("DATA ->\t\t%s\n", info.dataload);
        }			
}

void dump_udp(struct packet_info info, int data)
{
	int count = 1;

        printf("\nUDP Packet #%i\n", udp_cn++);
	printf("%u.%u.%u.%u:%i > %u.%u.%u.%u:%i\t\t(TTL:%i)\n",
		info.saddr[0],info.saddr[1],info.saddr[2],
		info.saddr[3], info.source, info.daddr[0],
                info.daddr[1], info.daddr[2], info.daddr[3],
                info.dest, info.ttl);
        if (data == 1) {
                printf("DATA ->\t\t%s\n", info.dataload);
        }
}

void dump_icmp(struct packet_info info)
{
	printf("\nICMP Packet\n");
        printf("%u.%u.%u.%u > %u.%u.%u.%u \ticmp : ",
                info.saddr[0],info.saddr[1],info.saddr[2],
                info.saddr[3], info.daddr[0], info.daddr[1], 
		info.daddr[2], info.daddr[3]);
        switch((info.type/256)) {
		case 0:
			printf("echo reply\t");
			break;
		case 3:
			printf("dest_unreach\t");
			break;
		case 4:
			printf("source quench\t");
			break;
		case 5:
			printf("redirect\t");
			break;
		case 8:
			printf("echo\t");
			break;
		case 11:
			printf("time exceeded\t");
			break;
		case 12:
			printf("parameter problem\t");
			break;
		case 13:
			printf("timestamp\t");
			break;
		case 14:
			printf("timestamp reply\t");
			break;
		case 15:
			printf("information\t");
			break;
		case 16:
			printf("information reply\t");
			break;
		case 17:
			printf("address mask\t");
			break;
		case 18:
			printf("address mask reply\t");
			break;
		default:
			printf("%i\t", info.type);
			break;
	}
	printf("(ttl:%i id:%i)\n", info.ttl, (info.id/256));
}

int main(int argc, char **argv)
{
	int snoop = 0, opt;
	struct packet_info pk_info;

	printf("\nSiMPLE LiNUX SNiFFER - v.0.3 - NO(C)1998 FuSyS\n\n");

	while ((opt = getopt(argc, argv, "v")) != EOF) {
                switch(opt)
                {
			case 'v':
				snoop=1;
				break;
			default:
				exit(1);
				break;
		}
	}

	signal(SIGINT, ethclose);
        signal(SIGTERM, ethclose);
	signal(SIGKILL, ethclose);
	signal(SIGQUIT, ethclose);

	sniff = init();
	while(1) {
		sniff_pk(sniff, &pk_info);
		switch(pk_info.protocol) {
			case IPPROTO_TCP:
				dump_tcp(pk_info, snoop);
				break;
			case IPPROTO_UDP:
				dump_udp(pk_info, snoop);
				break;
			case IPPROTO_ICMP:
				dump_icmp(pk_info);
				break;
			default:
				break;
		}
	}
}
